import random

import numpy as np

from moving_targets.learners import TensorflowMLP
from test.learners.test_learners import TestLearners


# TENSORFLOW IS IMPORTED LAZILY TO AVOID CONFLICTS WITH DEPENDENCIES TESTS

class TestTensorflowLearners(TestLearners):
    _EPOCHS: int = 2
    _UNITS: int = 64
    _BATCH_SIZE: int = 4
    _OPTIMIZER: str = 'rmsprop'
    _SHUFFLE: bool = True

    @classmethod
    def _random_state(cls):
        import tensorflow as tf
        random.seed(cls.SEED)
        np.random.seed(cls.SEED)
        tf.random.set_seed(cls.SEED)

    def _reference(self, learner, x, y, sample_weight) -> np.ndarray:
        learner.fit(x=x,
                    y=y,
                    sample_weight=sample_weight,
                    epochs=self._EPOCHS,
                    batch_size=self._BATCH_SIZE,
                    shuffle=self._SHUFFLE,
                    verbose=False)
        return learner.predict(x).squeeze()

    def test_regression_mlp(self):
        from tensorflow.keras.models import Sequential
        from tensorflow.keras.layers import Dense

        def ref():
            model = Sequential([Dense(self._UNITS, activation='relu'), Dense(1, activation=None)])
            model.compile(optimizer=self._OPTIMIZER, loss='mse')
            return model

        self._test(
            mt_learner=lambda: TensorflowMLP(
                loss='mse',
                optimizer=self._OPTIMIZER,
                hidden_units=[self._UNITS],
                epochs=self._EPOCHS,
                batch_size=self._BATCH_SIZE,
                shuffle=self._SHUFFLE,
                warm_start=1,  # set this to avoid call to clone_model() which messes up the random state
                verbose=False
            ),
            ref_learner=ref,
            classification=False
        )

    def test_classification_mlp(self):
        from tensorflow.keras.layers import Dense
        from tensorflow.keras.models import Sequential

        def ref():
            model = Sequential([Dense(self._UNITS, activation='relu'), Dense(1, activation='sigmoid')])
            model.compile(optimizer=self._OPTIMIZER, loss='binary_crossentropy')
            return model

        self._test(
            mt_learner=lambda: TensorflowMLP(
                loss='binary_crossentropy',
                output_activation='sigmoid',
                optimizer=self._OPTIMIZER,
                hidden_units=[self._UNITS],
                epochs=self._EPOCHS,
                batch_size=self._BATCH_SIZE,
                shuffle=self._SHUFFLE,
                warm_start=1,  # set this to avoid call to clone_model() which messes up the random state
                verbose=False
            ),
            ref_learner=ref,
            classification=True
        )



from typing import List, Any

from moving_targets.masters.backends import Backend, NumpyBackend
from test.masters.backends.test_backend import TestBackend


class TestNumpyBackend(TestBackend):
    @classmethod
    def _backend(cls) -> Backend:
        return NumpyBackend()

    @classmethod
    def _unsupported(cls) -> List[str]:
        return ['objectives', 'variables']

    @classmethod
    def _get_name(cls, variable: Any) -> str:
        return ''



import random

import numpy as np
from scipy.optimize import curve_fit

from moving_targets.learners import ScipyCurveFit
from test.learners.test_learners import TestLearners


class TestCurveLearners(TestLearners):
    @staticmethod
    def _curve(x, a, b):
        return a * x.sum(axis=1) + b

    @classmethod
    def _random_state(cls):
        random.seed(cls.SEED)
        np.random.seed(cls.SEED)

    def _reference(self, learner, x, y, sample_weight: np.ndarray) -> np.ndarray:
        if learner == 'scipy_curve_fit':
            a, b = curve_fit(f=self._curve, xdata=x, ydata=y)[0]
            return self._curve(x=x, a=a, b=b)
        else:
            raise AssertionError(f"Unknown reference learner '{learner}'")

    def test_scipy_curve_fit(self):
        self._test(mt_learner=lambda: ScipyCurveFit(curve=self._curve),
                   ref_learner=lambda: 'scipy_curve_fit',
                   classification=False)



from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

from moving_targets.metrics import MSE, MAE, R2
from test.metrics.test_metrics import TestMetrics


class TestRegressionMetrics(TestMetrics):
    @staticmethod
    def _data_generator(rng):
        y = rng.normal(size=TestRegressionMetrics.NUM_SAMPLES)
        p = rng.normal(size=TestRegressionMetrics.NUM_SAMPLES)
        return [], y, p

    def test_mae(self):
        self._test(data_generator=self._data_generator,
                   mt_metric=MSE(),
                   ref_metric=lambda x, y, p: mean_squared_error(y, p))

    def test_mse(self):
        self._test(data_generator=self._data_generator,
                   mt_metric=MAE(),
                   ref_metric=lambda x, y, p: mean_absolute_error(y, p))

    def test_r2(self):
        self._test(data_generator=self._data_generator,
                   mt_metric=R2(),
                   ref_metric=lambda x, y, p: r2_score(y, p))



from typing import Callable

import numpy as np

from moving_targets.metrics import Metric
from test.test_abstract import TestAbstract


class TestMetrics(TestAbstract):
    def _test(self, data_generator: Callable, mt_metric: Metric, ref_metric: Callable):
        """Performs the tests on the given data and checks the correctness of the metric wrt to a reference metric."""
        rng = np.random.default_rng(self.SEED)
        for i in range(self.NUM_TESTS):
            x, y, p = data_generator(rng)
            mt_value, ref_value = mt_metric(x, y, p), ref_metric(x, y, p)
            self.assertAlmostEqual(ref_value, mt_value, places=self.PLACES, msg=f'iteration: {i}')



import importlib.resources

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.model_selection import train_test_split

from moving_targets import MACS
from moving_targets.callbacks import DataLogger
from moving_targets.learners import LogisticRegression
from moving_targets.masters import ClassificationMaster
from moving_targets.masters.backends import GurobiBackend
from moving_targets.metrics import DIDI, CrossEntropy, Accuracy
# AS A FIRST STEP, WE NEED TO DEFINE OUR MASTER PROBLEM, WHICH IN THIS CASE WOULD BE THAT OF FAIR CLASSIFICATION
from moving_targets.util import probabilities


class FairClassification(ClassificationMaster):
    def __init__(self, protected, violation=0.2, backend='gurobi', loss='hd', alpha='harmonic', stats=False):
        # protected  : the name of the protected feature
        # violation  : the maximal accepted level of violation of the constraint.
        # backend    : the backend, which used to solve the master step
        # loss       : the loss function, which is used to compute the master objective
        # alpha      : the alpha optimizer, which is used to balance between the gradient term and the squared term
        # stats      : whether to include master statistics in the history object

        super().__init__(backend=backend, loss=loss, alpha=alpha, stats=stats, types='discrete', labelling=False)
        self.protected = protected
        self.violation = violation

    # here we define the problem formulation, i.e., variables and constraints
    def build(self, x, y, p):
        # retrieve model variables from the super method and, for compatibility between the binary/multiclass scenarios,
        # optionally transform 1d variables (representing a binary classification task) into a 2d matrix
        super_vars = super(FairClassification, self).build(x, y, p)
        variables = np.transpose([1 - super_vars, super_vars]) if super_vars.ndim == 1 else super_vars

        # as a first step, we need to compute the deviations between the average output for the total dataset and the
        # average output respectively to each protected class
        indicator_matrix = DIDI.get_indicator_matrix(x=x, protected=self.protected)
        groups, classes = len(indicator_matrix), len(np.unique(y))
        deviations = self.backend.add_continuous_variables(groups, classes, lb=0.0, name='deviations')
        # this is the average number of samples from the whole dataset <class[c]> as target class
        total_avg = [self.backend.mean(variables[:, c]) for c in range(classes)]
        for g, protected_group in enumerate(indicator_matrix):
            # this is the subset of the variables having <label> as protected feature (i.e., the protected group)
            protected_vars = variables[protected_group]
            if len(protected_vars) == 0:
                continue
            # this is the average number of samples within the protected group having <class[c]> as target
            protected_avg = [self.backend.mean(protected_vars[:, c]) for c in range(classes)]
            # eventually, the partial deviation is computed as the absolute value (which is linearized) of the
            # difference between the total average samples and the average samples within the protected group
            self.backend.add_constraints([deviations[g, c] >= total_avg[c] - protected_avg[c] for c in range(classes)])
            self.backend.add_constraints([deviations[g, c] >= protected_avg[c] - total_avg[c] for c in range(classes)])
        # finally, we compute the DIDI as the sum of this deviations, which is constrained to be lower or equal to the
        # given value (also, since we are computing the percentage DIDI, we need to scale for the original train_didi)
        didi = self.backend.sum(deviations)
        train_didi = DIDI.classification_didi(indicator_matrix=indicator_matrix, targets=y)
        self.backend.add_constraint(didi <= self.violation * train_didi)

        # return the original model variables at the end
        return super_vars


# THEN, WE MAY ADD A CUSTOM CALLBACK TO SEE HOW OUR TRAINING HAS PROCEEDED
class FairnessPlots(DataLogger):
    def __init__(self, protected, num_columns=4, **plt_kwargs):
        # protected   : the name of the protected feature
        # num_columns : the number of columns to display the subplots
        # plt_kwargs  : custom arguments to be passed to the 'matplotlib.pyplot.plot' function

        super().__init__()
        self.protected = protected
        self.num_columns = num_columns
        self.plt_kwargs = plt_kwargs

    def on_process_start(self, macs, x, y, val_data):
        # retrieve the subset of input features regarding the protected groups and store them in the inner 'data'
        # variable by replacing the input data which will not be useful during the plotting
        group = x[[c for c in x.columns if c.startswith(self.protected)]].values.squeeze().astype(int)
        self.data['group'] = group.argmax(axis=1) if group.ndim == 2 else group

    def on_training_end(self, macs, x, y, p, val_data):
        # store class targets instead of class probabilities
        super(FairnessPlots, self).on_training_end(macs, x, y, probabilities.get_classes(p), val_data)

    def on_process_end(self, macs, x, y, val_data):
        plt.figure(**self.plt_kwargs)
        num_rows = int(np.ceil(len(self.iterations) / self.num_columns))
        ax = None
        for it in self.iterations:
            ax = plt.subplot(num_rows, self.num_columns, it + 1, sharex=ax, sharey=ax)
            # we do not use count plot since we would like to show the percentage of predicted classes per group in
            # order to better see how moving targets affect the class balancing, thus we use instead a standard bar
            # plot to plot the bars with normalized value counts
            groups = self.data.rename(columns={f'p{it}': 'prediction'}).groupby('group')
            groups['prediction'].value_counts(normalize=True).mul(100).unstack().plot(kind='bar', stacked=True, ax=ax)
            ax.set(xlabel='group', ylabel='%')
            ax.set_title(f'iteration: {it}')
        plt.show()


if __name__ == '__main__':
    sns.set_style('whitegrid')
    sns.set_context('notebook')
    np.random.seed(0)

    # retrieve data and split train/test
    with importlib.resources.path('res', 'adult.csv') as filepath:
        df = pd.read_csv(filepath)
        x_df, y_df = df.drop('income', axis=1), df['income'].astype('category').cat.codes.values
        x_tr, x_ts, y_tr, y_ts = train_test_split(x_df, y_df, stratify=y_df, shuffle=True)

    # create a moving targets instance and fit it, then plot the training history
    # moreover, we pass a custom backend with a time limit
    cbs = [FairnessPlots(protected='race', figsize=(16, 9), tight_layout=True)]
    model = MACS(
        init_step='pretraining',
        learner=LogisticRegression(x_scaler='std', max_iter=10000),
        master=FairClassification(protected='race', backend=GurobiBackend(time_limit=10)),
        metrics=[Accuracy(), CrossEntropy(), DIDI(protected='race', classification=True, percentage=True)]
    )
    history = model.fit(x=x_tr, y=y_tr, iterations=10, val_data={'test': (x_ts, y_ts)}, callbacks=cbs, verbose=True)
    history.plot()



import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = u'Moving Targets'
copyright = u'2021, University of Bologna - DISI'
author = u'University of Bologna - DISI'

# The short X.Y version
version = u''
# The full version, including alpha/beta/rc tags
release = u''

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames. You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation for a list of supported languages.
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and directories to ignore when looking for source
# files. This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [u'_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme further.  For a list of options available
# for each theme, see the documentation.
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here, relative to this directory. They are
# copied after the builtin static files, so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names to template names.
# The default sidebars (for documents that don't match any pattern) are defined by theme itself.
# Builtin themes are using these templates by default:
#   ``['localtoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html']``.
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'MovingTargetsDoc'

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
#   (source start file, target name, title, author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'MovingTargets.tex', u'Moving Targets Documentation', u'University of Bologna - DISI', 'manual'),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
#   (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'movingtargets', u'Moving Targets Documentation', [author], 1)]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'MovingTargets', u'Moving Targets Documentation', author,
     'Moving Targets', 'One line description of project.', 'Miscellaneous'),
]

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'pandas': ('https://pandas.pydata.org/docs', None),
    'scikit-learn': ('https://scikit-learn.org/stable', None),
    'tensorflow': ('https://www.tensorflow.org/api_docs/python',
                   'https://raw.githubusercontent.com/GPflow/tensorflow-intersphinx/master/tf2_py_objects.inv')
}

# -- Options for to do extension ----------------------------------------------

# If true, `to do` and `to do list` produce output, else they produce nothing.
todo_include_todos = True



from typing import List, Any

from moving_targets.masters.backends import Backend, CvxpyBackend
from test.masters.backends.test_backend import TestBackend


class TestCvxpyBackend(TestBackend):
    @classmethod
    def _backend(cls) -> Backend:
        return CvxpyBackend()

    @classmethod
    def _unsupported(cls) -> List[str]:
        return ['is_greater', 'is_less', 'min', 'max', 'sqrt', 'var', 'cov', 'norm_0', 'norm_inf']

    @classmethod
    def _get_name(cls, variable: Any) -> str:
        return str(variable)



# Run this script from the origin folder as:
#   > "python document.py" in order to clean previous builds and generate new documentation
#   > "python document.py clean" in order to clean previous builds without generating new documentation

import os
import shutil
import sys

# clear previous documentation
shutil.rmtree('docs/html', ignore_errors=True)
shutil.rmtree('docs/source', ignore_errors=True)

# if clean is not passed, generate new documentation .rst with sphinx-apidoc for main module, then make html
if 'clean' not in sys.argv:
    os.mkdir('docs/source')
    os.system(f'sphinx-apidoc -e -T -M -d 8 -o .\\docs\\source .\\moving_targets')
    os.system('sphinx-build .\\docs .\\docs\\html')



from typing import Dict

import numpy as np
import pandas as pd

from moving_targets.metrics import ClassFrequenciesStd, MonotonicViolation, DIDI
from moving_targets.util import probabilities
from test.metrics.test_metrics import TestMetrics


class TestConstraintsMetrics(TestMetrics):
    DIDI_DATA: pd.DataFrame = pd.DataFrame.from_dict({
        'bin_protected': [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
        'multi_protected': [0, 0, 0, 1, 1, 1, 1, 2, 2, 2],
        'onehot_protected_0': [1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
        'onehot_protected_1': [0, 0, 0, 1, 1, 1, 1, 0, 0, 0],
        'onehot_protected_2': [0, 0, 0, 0, 0, 0, 0, 1, 1, 1],
        'reg_target': [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
        'bin_target': [1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
        'multi_target': [1, 2, 0, 0, 0, 0, 1, 1, 1, 2]
    })
    """Dataframe for DIDI tests: it con considers different protected features to test all the possible inputs and
    different target features to test all the possible tasks."""

    DIDI_RESULTS: Dict[str, Dict[str, float]] = {
        'bin_protected': {
            'reg_target': 0.5,
            'bin_target': 0.8,
            'multi_target': 0.8
        },
        'multi_protected': {
            'reg_target': 0.7,
            'bin_target': 1.6333,
            'multi_target': 1.7666
        },
        'onehot_protected': {
            'reg_target': 0.7,
            'bin_target': 1.6333,
            'multi_target': 1.7666
        }
    }
    """Dictionary of results indexed by protected class type and target class type."""

    def test_class_frequencies_std(self):
        def data_generator(rng):
            # create a dictionary of data counts per class
            counts = {c: 1 + rng.integers(self.NUM_SAMPLES) for c in range(self.NUM_CLASSES)}
            values = np.array([v for v in counts.values()]) / np.sum([v for v in counts.values()])
            # create vector of classes and shuffle it, then obtain fake probabilities due to metric compatibility
            classes = np.concatenate([c * np.ones(n, dtype=int) for c, n in counts.items()])
            rng.shuffle(classes)
            # store the class value counts as y term to be used as ground truth
            return [], values, probabilities.get_onehot(vector=classes)

        self._test(data_generator=data_generator,
                   mt_metric=ClassFrequenciesStd(),
                   ref_metric=lambda x, y, p: y.std())

    def test_monotonic_violation(self):
        def diff(v):
            # computes pairwise differences between data points
            return np.array([[vi - vj for vj in v] for vi in v])

        def mono(v):
            # we consider ascending order, thus the expected monotonicities are computed as the sign of the differences
            return np.sign(diff(v))

        def data_generator(rng):
            # no need for y data
            x = rng.random(self.NUM_SAMPLES)
            p = rng.random(self.NUM_SAMPLES)
            # violations are computed by getting only positive differences between pairwise predictions and then
            # masking the values having expected decreasing monotonicity (so that the increasing dual is not counted)
            violations = diff(p)[mono(x) == -1]
            violations[violations < 1e-3] = 0.0
            # store the violations as y term to be used as ground truth
            return x, violations, p

        # test average
        self._test(data_generator=data_generator,
                   mt_metric=MonotonicViolation(monotonicities_fn=mono, aggregation='average'),
                   ref_metric=lambda x, y, p: y.mean())

        # test percentage
        self._test(data_generator=data_generator,
                   mt_metric=MonotonicViolation(monotonicities_fn=mono, aggregation='percentage'),
                   ref_metric=lambda x, y, p: np.sign(y).mean())

        # test feasibility
        self._test(data_generator=data_generator,
                   mt_metric=MonotonicViolation(monotonicities_fn=mono, aggregation='feasible'),
                   ref_metric=lambda x, y, p: float(np.all(y == 0)))

    def test_didi(self):
        for protected, target_dict in self.DIDI_RESULTS.items():
            for target, didi in target_dict.items():
                x = self.DIDI_DATA.drop(columns=[target])
                y = self.DIDI_DATA[target]
                if 'reg' in target:
                    p = y
                    didi_abs = DIDI(classification=False, protected=protected, percentage=False)
                    didi_per = DIDI(classification=False, protected=protected, percentage=True)
                else:
                    p = probabilities.get_onehot(vector=y.values)
                    didi_abs = DIDI(classification=True, protected=protected, percentage=False)
                    didi_per = DIDI(classification=True, protected=protected, percentage=True)
                msg = f"p: '{protected}', t: '{target}'"
                metric_didi_abs, metric_didi_per = didi_abs(x=x, y=y, p=p), didi_per(x=x, y=y, p=p)
                self.assertAlmostEqual(didi, metric_didi_abs, places=self.PLACES, msg=msg)
                self.assertAlmostEqual(1.0, metric_didi_per, places=self.PLACES, msg=msg)



from typing import Callable

import numpy as np
import pandas as pd

from moving_targets.util.errors import not_implemented_message
from test.test_abstract import TestAbstract


class TestLearners(TestAbstract):
    @classmethod
    def _random_state(cls):
        """Defines the random seeds to be fixed before calling the moving targets learner and the reference learner."""
        raise NotImplementedError(not_implemented_message(name='_random_state', static=True))

    def _reference(self, learner, x, y, sample_weight: np.ndarray) -> np.ndarray:
        """Implements the strategy to fit the reference learner and retrieve its predictions."""
        raise NotImplementedError(not_implemented_message(name='_reference'))

    def _test(self, mt_learner: Callable, ref_learner: Callable, classification: bool):
        """Performs the tests on the given data and checks the correctness of the learner wrt to a reference learner."""
        rng = np.random.default_rng(self.SEED)
        for weights in [True, False]:
            for i in range(self.NUM_TESTS):
                # generate data
                x = rng.random((self.NUM_SAMPLES, self.NUM_FEATURES))
                y = rng.integers(0, 2, self.NUM_SAMPLES) if classification else rng.random(self.NUM_SAMPLES)
                sample_weight = rng.random(self.NUM_SAMPLES) if weights else None
                # fit and predict using the moving targets learner
                self._random_state()
                mt_pred = mt_learner().fit(x, y, sample_weight=sample_weight).predict(x)
                # fit and predict using the reference learner
                self._random_state()
                ref_pred = self._reference(ref_learner(), x, y, sample_weight)
                # check correctness
                df = pd.DataFrame.from_dict({'y': y, 'mt': mt_pred, 'ref': ref_pred})
                diff = np.abs(mt_pred - ref_pred)
                msg = f'iteration: {i}, weights: {weights}, diff: {diff.max()} (sample {diff.argmax()})\n\n\n{df}'
                self.assertTrue(np.allclose(mt_pred, ref_pred, atol=10 ** -self.PLACES), msg=msg)



import inspect
from typing import Dict, Optional, Union, List, Set, Any

import moving_targets
from moving_targets.learners import Learner
from moving_targets.masters import Master
from moving_targets.masters.backends import Backend
from moving_targets.masters.losses import Loss
from moving_targets.masters.optimizers import Optimizer
from moving_targets.metrics import Metric
from moving_targets.util.scalers import Scaler
from test.test_abstract import TestAbstract


class TestSignatures(TestAbstract):
    MACS_SIGNATURE: Dict[str, Any] = {
        'master': Master,
        'learner': Learner,
        'init_step': (str, 'pretraining'),
        'metrics': (List[Metric], ()),
        'stats': (Union[bool, List[str]], False)
    }

    LEARNER_SIGNATURE: Dict[str, Any] = {
        'mask': (Optional[float], None),
        'x_scaler': (Union[None, Scaler, str], None),
        'y_scaler': (Union[None, Scaler, str], None),
        'stats': (Union[bool, List[str]], False)
    }

    MASTER_SIGNATURE: Dict[str, Any] = {
        'backend': Union[str, Backend],
        'loss': Union[str, Loss],
        'alpha': Union[str, float, Optimizer],
        'mask': (Optional[float], None),
        'x_scaler': (Union[None, Scaler, str], None),
        'y_scaler': (Union[None, Scaler, str], None),
        'stats': (Union[bool, List[str]], False)
    }

    LOSS_SIGNATURE: Dict[str, Any] = {
        'binary': bool,
        'name': str
    }

    def _test(self, module, signature: Dict[str, Any], excluded: Optional[Set[str]] = None):
        """
        :param module:
            The module to test.

        :param signature:
            A dictionary of annotations (and optional default values) indexed by parameters to match in the signature.

        :param excluded:
            A set of objects names to exclude from the check.
        """
        excluded = excluded or set()
        for name, obj in inspect.getmembers(module):
            if name not in excluded and inspect.isclass(obj):
                params = inspect.signature(obj.__init__).parameters
                for param, ann in signature.items():
                    self.assertIn(param, params, msg=f'\n"{name}" has no "{param}" parameter')
                    if isinstance(ann, tuple):
                        ann, dfl = ann
                        self.assertEqual(dfl, params[param].default, msg=f'"{name}" has wrong "{param}" default value')
                    self.assertEqual(ann, params[param].annotation, msg=f'"{name}" has wrong "{param}" annotation')

    def test_macs(self):
        self._test(moving_targets, self.MACS_SIGNATURE)

    def test_learners(self):
        # exclude the learner interface since it does not have default values
        self._test(moving_targets.learners, self.LEARNER_SIGNATURE, excluded={'Learner'})

    def test_masters(self):
        self._test(moving_targets.masters, self.MASTER_SIGNATURE)



from typing import Optional

import numpy as np
import pandas as pd

from moving_targets.util.masking import get_mask, mask_data
from test.test_abstract import TestAbstract


class TestMasking(TestAbstract):
    REF_1D = [0., 1., 2., 3., 4., 5.]
    REF_2D = [[i] * TestAbstract.NUM_CLASSES for i in REF_1D]
    MASK = np.array([True, True, False, True, False, False])

    def _test_get(self, ref: np.ndarray, mask: Optional[float], all_columns: bool, should_fail: bool):
        ref_mask = [True] * len(self.MASK) if mask is None else list(self.MASK)
        act_mask = list(get_mask(reference=ref, mask=mask, all_columns=all_columns))
        is_equal = np.equal(ref_mask, act_mask).all()
        msg = f"\nMasks {ref_mask} and {act_mask} are "
        if should_fail:
            self.assertFalse(is_equal, msg=msg + "not expected to be equal, but they are.")
        else:
            self.assertTrue(is_equal, msg=msg + "expected to be equal, but they are not.")

    def test_get_none_1d(self):
        self._test_get(ref=np.array(self.REF_1D), mask=None, all_columns=True, should_fail=False)
        self._test_get(ref=np.array(self.REF_1D), mask=None, all_columns=False, should_fail=False)

    def test_get_nan_1d(self):
        ref = np.where(self.MASK, self.REF_1D, np.nan)
        self._test_get(ref=ref, mask=np.nan, all_columns=True, should_fail=False)
        self._test_get(ref=ref, mask=np.nan, all_columns=False, should_fail=False)

    def test_get_val_1d(self):
        ref = np.where(self.MASK, self.REF_1D, -1)
        self._test_get(ref=ref, mask=-1, all_columns=True, should_fail=False)
        self._test_get(ref=ref, mask=-1, all_columns=True, should_fail=False)

    def test_get_none_2d(self):
        self._test_get(ref=np.array(self.REF_2D), mask=None, all_columns=True, should_fail=False)
        self._test_get(ref=np.array(self.REF_2D), mask=None, all_columns=False, should_fail=False)

    def test_get_nan_all(self):
        ref = np.array(self.REF_2D)
        for index in np.arange(len(self.MASK))[~self.MASK]:
            ref[index, :] = np.nan
        self._test_get(ref=ref, mask=np.nan, all_columns=True, should_fail=False)
        self._test_get(ref=ref, mask=np.nan, all_columns=False, should_fail=False)

    def test_get_val_all(self):
        ref = np.array(self.REF_2D)
        for index in np.arange(len(self.MASK))[~self.MASK]:
            ref[index, :] = -1
        self._test_get(ref=ref, mask=-1, all_columns=True, should_fail=False)
        self._test_get(ref=ref, mask=-1, all_columns=False, should_fail=False)

    def test_get_nan_any(self):
        rng = np.random.default_rng(self.SEED)
        ref = np.array(self.REF_2D)
        for index in np.arange(len(self.MASK))[~self.MASK]:
            ref[index, rng.integers(0, self.NUM_CLASSES, size=1)] = np.nan
        self._test_get(ref=ref, mask=np.nan, all_columns=False, should_fail=False)
        self._test_get(ref=ref, mask=np.nan, all_columns=True, should_fail=True)

    def test_get_val_any(self):
        rng = np.random.default_rng(self.SEED)
        ref = np.array(self.REF_2D)
        for index in np.arange(len(self.MASK))[~self.MASK]:
            ref[index, rng.integers(0, self.NUM_CLASSES, size=1)] = -1
        self._test_get(ref=ref, mask=-1, all_columns=False, should_fail=False)
        self._test_get(ref=ref, mask=-1, all_columns=True, should_fail=True)

    def test_masking(self):
        masked = np.array(self.REF_1D)[self.MASK]
        vectors = [
            np.array(self.REF_1D),
            pd.Series(self.REF_1D),
            pd.DataFrame(self.REF_1D),
            np.array([self.REF_1D] * self.NUM_FEATURES).T
        ]
        ref_vec = [masked, pd.Series(masked), pd.DataFrame(masked), np.array([masked] * self.NUM_FEATURES).T]
        act_vec = mask_data(*vectors, mask=self.MASK)
        for ref, act in zip(ref_vec, act_vec):
            self.assertIsInstance(ref, type(act))
            # deal with matrices by adapting all the other arrays to bi-dimensional shape
            ref, act = np.atleast_2d(ref), np.atleast_2d(act)
            for r, a in zip(ref, act):
                self.assertEqual(list(r), list(a))



from typing import Callable

import numpy as np

from moving_targets.masters import optimizers
from test.test_abstract import TestAbstract


class TestOptimizers(TestAbstract):

    def _test(self, policy: Callable, optimizer: str, **opt_args):
        """Checks whether the optimizer, if instantiated with the given parameters, behaves as the reference policy."""
        dummy = np.zeros(1)
        history = []
        optimizer = optimizers.aliases[optimizer](**opt_args)
        for i in np.arange(self.NUM_TESTS):
            ref_value = policy(it=i)
            step_value = optimizer(x=dummy, y=dummy, p=dummy)
            self.assertAlmostEqual(step_value, ref_value, places=self.PLACES)
            get_value = optimizer.value
            self.assertAlmostEqual(get_value, ref_value, places=self.PLACES)
            history.append(ref_value)
        msg = f"Reference List: {history}\nActual List: {optimizer.history}"
        self.assertTrue(np.allclose(history, optimizer.history, atol=10 ** -self.PLACES), msg=msg)

    def test_constant_default(self):
        self._test(policy=lambda it: 1, optimizer='constant')

    def test_constant_base(self):
        self._test(policy=lambda it: 2, optimizer='constant', base=2)

    def test_harmonic_default(self):
        self._test(policy=lambda it: 1 / (it + 1), optimizer='harmonic')

    def test_harmonic_base(self):
        self._test(policy=lambda it: 2 / (it + 1), optimizer='harmonic', base=2)

    def test_geometric_default(self):
        self._test(policy=lambda it: 1 / (2 ** it), optimizer='geometric')

    def test_geometric_base(self):
        self._test(policy=lambda it: 2 / (2 ** it), optimizer='geometric', base=2)

    def test_geometric_slope(self):
        self._test(policy=lambda it: 1 / (3 ** it), optimizer='geometric', slope=3)



import random

import numpy as np
import torch

from moving_targets.learners import TorchMLP
from test.learners.test_learners import TestLearners


# TORCH IS IMPORTED LAZILY TO AVOID CONFLICTS WITH DEPENDENCIES TESTS

class TestTorchLearners(TestLearners):
    _EPOCHS: int = 2
    _UNITS: int = 2
    _BATCH_SIZE: int = 4
    _OPTIMIZER: str = 'RMSprop'
    _SHUFFLE: bool = True

    @classmethod
    def _random_state(cls):
        import torch
        random.seed(cls.SEED)
        np.random.seed(cls.SEED)
        torch.manual_seed(cls.SEED)

    def _reference(self, learner, x, y, sample_weight) -> np.ndarray:
        from moving_targets.learners import TorchLearner
        from torch.nn import Sequential, Linear, ReLU, Sigmoid, MSELoss, BCELoss
        from torch.optim import RMSprop
        from torch.utils.data import DataLoader

        if learner == 'regression_mlp':
            learner = Sequential(Linear(self.NUM_FEATURES, self._UNITS), ReLU(), Linear(self._UNITS, 1))
            loss = MSELoss()
        elif learner == 'classification_mlp':
            learner = Sequential(Linear(self.NUM_FEATURES, self._UNITS), ReLU(), Linear(self._UNITS, 1), Sigmoid())
            loss = BCELoss()
        else:
            raise AssertionError(f"Unknown reference learner '{learner}'")

        if self._OPTIMIZER == 'RMSprop':
            optimizer = RMSprop(learner.parameters())
        else:
            raise AssertionError(f"Unknown optimizer '{self._OPTIMIZER}'")

        # noinspection PyTypeChecker
        loader = DataLoader(dataset=TorchLearner._Dataset(x=x, y=y), batch_size=self._BATCH_SIZE, shuffle=self._SHUFFLE)
        learner.train()
        for i in range(self._EPOCHS):
            for inp, out in loader:
                optimizer.zero_grad()
                pred = learner(inp)
                loss(pred, out).backward()
                optimizer.step()
        learner.eval()
        return learner(torch.tensor(np.array(x), dtype=torch.float32)).detach().numpy().squeeze()

    def test_regression_mlp(self):
        self._test(
            mt_learner=lambda: TorchMLP(
                loss='MSELoss',
                optimizer=self._OPTIMIZER,
                input_units=self.NUM_FEATURES,
                hidden_units=[self._UNITS],
                iterations=self._EPOCHS,
                batch_size=self._BATCH_SIZE,
                shuffle=self._SHUFFLE,
                verbose=False
            ),
            ref_learner=lambda: 'regression_mlp',
            classification=False)

    def test_classification_mlp(self):
        self._test(
            mt_learner=lambda: TorchMLP(
                loss='BCELoss',
                output_activation='Sigmoid',
                optimizer=self._OPTIMIZER,
                input_units=self.NUM_FEATURES,
                hidden_units=[self._UNITS],
                iterations=self._EPOCHS,
                batch_size=self._BATCH_SIZE,
                shuffle=self._SHUFFLE,
                verbose=False
            ),
            ref_learner=lambda: 'classification_mlp',
            classification=True
        )



import os
from typing import List, Any

import pytest

from moving_targets.masters.backends import Backend, GurobiBackend
from test.masters.backends.test_backend import TestBackend


@pytest.mark.skipif(os.getenv('GITHUB_ACTIONS') == 'true', reason='No solver in Github Actions')
class TestGurobiBackend(TestBackend):
    @classmethod
    def _backend(cls) -> Backend:
        return GurobiBackend()

    @classmethod
    def _unsupported(cls) -> List[str]:
        return []

    @classmethod
    def _get_name(cls, variable: Any) -> str:
        return str(variable)[12:-13].rstrip(']').replace('[', '_').replace(',', '_')



import os
from typing import Optional, Tuple

import numpy as np
import pytest

from moving_targets.masters import losses
from moving_targets.masters.backends import GurobiBackend
from moving_targets.util import probabilities
from test.test_abstract import TestAbstract


@pytest.mark.skipif(os.getenv('GITHUB_ACTIONS') == 'true', reason='No solver in Github Actions')
class TestLosses(TestAbstract):
    @staticmethod
    def _normalize(array: np.ndarray) -> np.ndarray:
        """Normalizes the values of an array so that each of its rows sum up to one."""
        return array if array.ndim == 1 else np.transpose(array.transpose() / array.sum(axis=1))

    @staticmethod
    def _ref_loss(loss: str,
                  values: np.ndarray,
                  targets: np.ndarray,
                  predictions: np.ndarray,
                  weights: Optional[np.ndarray] = None) -> Tuple[float, float]:
        """Implements the correct loss definition based on the derivative of the given reference loss."""
        import tensorflow as tf
        import tensorflow.keras.losses as ls
        # handle inputs
        v = tf.constant(values.reshape((len(values), -1)), dtype=tf.float32)
        t = tf.constant(targets.reshape((len(targets), -1)), dtype=tf.float32)
        p = tf.Variable(predictions.reshape((len(predictions), -1)), dtype=tf.float32)
        w = None if weights is None else tf.constant(len(weights) * weights / weights.sum(), dtype=tf.float32)
        # handle loss
        if loss == 'mae':
            loss = ls.MeanAbsoluteError(reduction='none')
        elif loss == 'mse':
            loss = ls.MeanSquaredError(reduction='none')
        elif loss == 'bce':
            loss = ls.BinaryCrossentropy(reduction='none')
        elif loss == 'cce':
            loss = ls.CategoricalCrossentropy(reduction='none')
        elif loss in ['bhd', 'chd']:
            def _hamming_distance(y_true, y_pred, sample_weight):
                partial_losses = tf.reduce_mean(y_true * (1 - y_pred) + (1 - y_true) * y_pred, axis=1)
                return partial_losses if sample_weight is None else sample_weight * partial_losses

            loss = _hamming_distance
        else:
            raise AssertionError(f"Unknown loss {loss}")
        # handle differentiation
        with tf.GradientTape() as tape:
            loss = loss(t, p, sample_weight=w)
            nabla = tape.gradient(loss, p)
        nabla_term = tf.reduce_mean(tf.reduce_sum(nabla * (v - p), axis=1))
        squared_term = ls.MeanSquaredError().__call__(v, p, sample_weight=w)
        return nabla_term.numpy(), squared_term.numpy()

    def _test(self, loss: str, task: str, cls: Optional[int], weights: bool, **loss_args):
        """Checks that the given moving targets loss behaves as the reference one with respect to the given task (which
        is used to account for the postprocessing needed for the predictions), the given number of classes (in case of
        classification tasks), and the given vector of sample weights."""
        try:
            rng = np.random.default_rng(self.SEED)
            backend = GurobiBackend()
            size = (self.NUM_SAMPLES,) if cls is None or cls == 2 else (self.NUM_SAMPLES, cls)
            kind = 'binary' if task in ['indicator', 'probability'] else 'continuous'
            mt_loss = losses.aliases[loss](**loss_args)
            for i in range(self.NUM_TESTS):
                backend.build()
                # assign alpha, predictions, and sample weights
                alpha = rng.uniform(0, 1)
                predictions = self._normalize(rng.uniform(0, 1, size=size))
                sample_weight = rng.uniform(size=self.NUM_SAMPLES) if weights else None
                if kind == 'binary':
                    # for binary reference values and targets, create a vector of classes and then binarize them
                    values = probabilities.get_onehot(vector=rng.integers(0, cls, size=self.NUM_SAMPLES), classes=cls)
                    targets = probabilities.get_onehot(vector=rng.integers(0, cls, size=self.NUM_SAMPLES), classes=cls)
                else:
                    # for continuous reference values and targets, create a vector of outputs and then normalize
                    values = self._normalize(rng.uniform(0, 1, size=size))
                    targets = self._normalize(rng.uniform(0, 1, size=size))
                # create constant model variables from values then compute the backend objective
                variables = backend.add_constants(values, vtype=kind, name='var')
                mt_nabla, mt_squared = mt_loss(
                    backend=backend,
                    variables=variables,
                    targets=targets,
                    predictions=predictions,
                    sample_weight=sample_weight
                )
                mt_value = backend.minimize(alpha * mt_nabla + mt_squared).solve().get_objective()
                mt_nabla, mt_squared = backend.get_value(mt_nabla), backend.get_value(mt_squared)
                backend.clear()
                # optionally post-process the values then compute the reference objective
                if task == 'indicator':
                    labelling = loss_args.get('labelling') or False
                    values = probabilities.get_classes(values, labelling=labelling)
                    targets = probabilities.get_classes(targets, labelling=labelling)
                    predictions = probabilities.get_classes(predictions, labelling=labelling)
                    if not labelling:
                        values = probabilities.get_onehot(values, classes=cls)
                        targets = probabilities.get_onehot(targets, classes=cls)
                        predictions = probabilities.get_onehot(predictions, classes=cls)
                ref_nabla, ref_squared = self._ref_loss(
                    loss=loss,
                    values=values,
                    targets=targets,
                    predictions=predictions,
                    weights=sample_weight
                )
                ref_value = alpha * ref_nabla + ref_squared
                # compare moving targets losses with the reference ones
                self.assertAlmostEqual(mt_nabla, ref_nabla, places=self.PLACES, msg=f'Error in nabla, it: {i}')
                self.assertAlmostEqual(mt_squared, ref_squared, places=self.PLACES, msg=f'Error in squared, it: {i}')
                self.assertAlmostEqual(mt_value, ref_value, places=self.PLACES, msg=f'Error in loss, it: {i}')
        except NotImplementedError:
            # the abstract testcase will be executed as well, thus if we check whether we are running one of its tests
            self.assertTrue(self.__class__.__name__ == 'TestBackend')

    def test_mae(self):
        self._test(loss='mae', task='regression', cls=None, weights=False)

    def test_mae_weights(self):
        self._test(loss='mae', task='regression', cls=None, weights=True)

    def test_binary_mae(self):
        self._test(loss='mae', task='probability', cls=self.NUM_CLASSES, weights=False, binary=True)

    def test_binary_mae_weights(self):
        self._test(loss='mae', task='probability', cls=self.NUM_CLASSES, weights=True, binary=True)

    def test_mse(self):
        self._test(loss='mse', task='regression', cls=None, weights=False)

    def test_mse_weights(self):
        self._test(loss='mse', task='regression', cls=None, weights=True)

    def test_binary_mse(self):
        self._test(loss='mse', task='probability', cls=self.NUM_CLASSES, weights=False, binary=True)

    def test_binary_mse_weights(self):
        self._test(loss='mse', task='probability', cls=self.NUM_CLASSES, weights=True, binary=True)

    def test_bh(self):
        self._test(loss='bhd', task='indicator', cls=2, weights=False)

    def test_bh_weights(self):
        self._test(loss='bhd', task='indicator', cls=2, weights=True)

    def test_labelling_bh(self):
        self._test(loss='bhd', task='indicator', cls=2, weights=False, labelling=True)

    def test_labelling_bh_weights(self):
        self._test(loss='bhd', task='indicator', cls=2, weights=True, labelling=True)

    def test_ch(self):
        self._test(loss='chd', task='indicator', cls=self.NUM_CLASSES, weights=False)

    def test_ch_weights(self):
        self._test(loss='chd', task='indicator', cls=self.NUM_CLASSES, weights=True)

    def test_labelling_ch(self):
        self._test(loss='chd', task='indicator', cls=self.NUM_CLASSES, weights=False, labelling=True)

    def test_labelling_ch_weights(self):
        self._test(loss='chd', task='indicator', cls=self.NUM_CLASSES, weights=True, labelling=True)

    def test_bce(self):
        self._test(loss='bce', task='probability', cls=2, weights=False)

    def test_bce_weights(self):
        self._test(loss='bce', task='probability', cls=2, weights=True)

    def test_binary_bce(self):
        self._test(loss='bce', task='probability', cls=2, weights=False, binary=True)

    def test_binary_bce_weights(self):
        self._test(loss='bce', task='probability', cls=2, weights=True, binary=True)

    def test_cce(self):
        self._test(loss='cce', task='probability', cls=self.NUM_CLASSES, weights=False)

    def test_cce_weights(self):
        self._test(loss='cce', task='probability', cls=self.NUM_CLASSES, weights=True)

    def test_binary_cce(self):
        self._test(loss='cce', task='probability', cls=self.NUM_CLASSES, weights=False, binary=True)

    def test_binary_cce_weights(self):
        self._test(loss='cce', task='probability', cls=self.NUM_CLASSES, weights=True, binary=True)



import os
import sys
import unittest


class TestAbstract(unittest.TestCase):
    SEED: int = 42
    """The chosen random seed."""

    NUM_TESTS: int = 10
    """The number of tests carried out."""

    NUM_SAMPLES: int = 10
    """The number of data points."""

    NUM_FEATURES: int = 5
    """The number of input features."""

    NUM_CLASSES: int = 3
    """The number of class labels for multi-class classification tests."""

    PLACES: int = 3
    """The number of digits passed to the `assertAlmostEqual()` method."""

    @staticmethod
    def get_relative_path(*path: str) -> str:
        # paths may have different positions based on the environment in which the function is called
        # we look for the path of the project, i.e., the one that ends with the project name
        root = None
        for p in sys.path:
            if p.endswith('Moving Targets'):
                root = p
                break
        return os.sep.join((root,) + path)



import os
from typing import Union, Optional

import numpy as np
import pytest

from moving_targets import MACS
from moving_targets.learners import Learner
from moving_targets.masters import ClassificationMaster, RegressionMaster
from moving_targets.masters.backends import GurobiBackend
from moving_targets.util import probabilities
from test.test_abstract import TestAbstract


@pytest.mark.skipif(os.getenv('GITHUB_ACTIONS') == 'true', reason='No solver in Github Actions')
class TestMasters(TestAbstract):
    class DummyLearner(Learner):
        def __init__(self, regression: bool):
            super(TestMasters.DummyLearner, self).__init__(mask=None, x_scaler=None, y_scaler=None, stats=False)
            self.regression: bool = regression
            self.p: Optional[np.ndarray] = None

        def _fit(self, x, y, sample_weight=None):
            self.p = y

        def _predict(self, x):
            return self.p if self.regression or self.p.ndim == 2 else probabilities.get_onehot(self.p.astype(int))

    def _test(self, task: Union[int, str], loss: str, types: Optional[str] = None):
        """Checks that the masters implementations behave correctly with respect to the given parameters."""
        rng = np.random.default_rng(self.SEED)
        x = rng.random((self.NUM_SAMPLES, self.NUM_FEATURES))
        # handle types
        if types == 'discrete':
            binary, classes = True, True
        elif types == 'discretized':
            binary, classes = False, True
        elif types == 'continuous':
            binary, classes = False, False
        else:
            binary, classes = None, None
        # handle outputs
        if task == 'regression':
            p = y = rng.random(self.NUM_SAMPLES)
            master = RegressionMaster(backend=GurobiBackend(), loss=loss)
        else:
            if task == 'binary':
                y = rng.random(self.NUM_SAMPLES).round().astype(int)
                p = y if classes else y.astype(float)
                master = ClassificationMaster(backend=GurobiBackend(), loss=loss, types=types)
            elif task == 'multiclass':
                y = rng.random((self.NUM_SAMPLES, self.NUM_CLASSES)).argmax(axis=1)
                p = y if classes else probabilities.get_onehot(y).astype(float)
                master = ClassificationMaster(backend=GurobiBackend(), loss=loss, types=types)
            elif task == 'multilabel':
                y = rng.random((self.NUM_SAMPLES, self.NUM_CLASSES)).round().astype(int)
                p = y if classes else y.astype(float)
                master = ClassificationMaster(backend=GurobiBackend(), loss=loss, types=types, labelling=True)
            else:
                raise AssertionError(f"Unsupported task '{task}'")
            self.assertEqual(master.binary, binary)
            self.assertEqual(master.classes, classes)
        # check correctness of targets
        learner = self.DummyLearner(regression=task == 'regression')
        MACS(learner=learner, master=master).fit(x, y, iterations=1)
        self.assertEqual(learner.p.shape, p.shape)
        self.assertEqual(learner.p.dtype, p.dtype)
        msg = f'Loss: {loss}\n\nRef:\n{learner.p}\n\nAct:\n{p}'
        self.assertTrue(np.allclose(learner.p, p, atol=10 ** -self.PLACES), msg=msg)

    def test_regression(self):
        self._test(task='regression', loss='mae')
        self._test(task='regression', loss='mse')

    def test_binary_discrete(self):
        self._test(task='binary', loss='hd', types='discrete')
        self._test(task='binary', loss='ce', types='discrete')
        self._test(task='binary', loss='mae', types='discrete')
        self._test(task='binary', loss='mse', types='discrete')

    def test_binary_discretized(self):
        self._test(task='binary', loss='ce', types='discretized')
        self._test(task='binary', loss='mae', types='discretized')
        self._test(task='binary', loss='mse', types='discretized')

    def test_binary_continuous(self):
        self._test(task='binary', loss='ce', types='continuous')
        self._test(task='binary', loss='mae', types='continuous')
        self._test(task='binary', loss='mse', types='continuous')

    def test_multiclass_discrete(self):
        self._test(task='multiclass', loss='hd', types='discrete')
        self._test(task='multiclass', loss='ce', types='discrete')
        self._test(task='multiclass', loss='mae', types='discrete')
        self._test(task='multiclass', loss='mse', types='discrete')

    def test_multiclass_discretized(self):
        self._test(task='multiclass', loss='ce', types='discretized')
        self._test(task='multiclass', loss='mae', types='discretized')
        self._test(task='multiclass', loss='mse', types='discretized')

    def test_multiclass_continuous(self):
        self._test(task='multiclass', loss='ce', types='continuous')
        self._test(task='multiclass', loss='mae', types='continuous')
        self._test(task='multiclass', loss='mse', types='continuous')

    def test_multilabel_discrete(self):
        self._test(task='multilabel', loss='hd', types='discrete')
        self._test(task='multilabel', loss='ce', types='discrete')
        self._test(task='multilabel', loss='mae', types='discrete')
        self._test(task='multilabel', loss='mse', types='discrete')

    def test_multilabel_discretized(self):
        self._test(task='multilabel', loss='ce', types='discretized')
        self._test(task='multilabel', loss='mae', types='discretized')
        self._test(task='multilabel', loss='mse', types='discretized')

    def test_multilabel_continuous(self):
        self._test(task='multilabel', loss='ce', types='continuous')
        self._test(task='multilabel', loss='mae', types='continuous')
        self._test(task='multilabel', loss='mse', types='continuous')



from typing import Callable

import numpy as np
from sklearn.preprocessing import LabelBinarizer

from moving_targets.util.probabilities import get_classes, get_onehot
from test.test_abstract import TestAbstract


class TestProbabilities(TestAbstract):
    BINARY_PROBABILITIES = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    BINARY_CLASSES = BINARY_PROBABILITIES.round().astype(int)
    BINARY_ONEHOT = np.transpose([1 - BINARY_CLASSES, BINARY_CLASSES])
    """1d vector of binary probabilities, 1d vector of binary classes, and 2d vector of binary classes, respectively."""

    MULTICLASS_PROBABILITIES = np.array([
        [0.7, 0.2, 0.1],
        [0.5, 0.4, 0.1],
        [0.4, 0.3, 0.3],
        [0.2, 0.6, 0.2],
        [0.3, 0.5, 0.2],
        [0.0, 1.0, 0.0],
        [0.1, 0.5, 0.4],
        [0.1, 0.4, 0.5],
        [0.4, 0.1, 0.5],
        [0.2, 0.2, 0.6]
    ])
    MULTICLASS_CLASSES = MULTICLASS_PROBABILITIES.argmax(axis=1)
    MULTICLASS_ONEHOT = LabelBinarizer().fit_transform(MULTICLASS_CLASSES)
    """2d vector of probabilities, 1d vector of classes, and 2d vector of onehot classes, respectively."""

    MULTILABEL_PROBABILITIES = np.array([
        [0.7, 0.4, 0.6],
        [0.5, 0.9, 0.1],
        [0.2, 0.3, 0.3],
        [0.4, 0.6, 0.2],
        [0.3, 0.5, 0.5],
        [0.0, 1.0, 0.0],
        [0.1, 0.1, 0.8],
        [0.1, 0.9, 0.5],
        [0.7, 0.9, 0.6],
        [0.5, 0.4, 0.6]
    ])
    MULTILABEL_ONEHOT = MULTILABEL_PROBABILITIES.round().astype(int)
    """2d vector of label probabilities, and 2d vector of onehot labels, respectively."""

    CATEGORIES_CLASSES = MULTICLASS_CLASSES + 1
    CATEGORIES_ONEHOT = np.array([
        [0, 1, 0, 0, 0],
        [0, 1, 0, 0, 0],
        [0, 1, 0, 0, 0],
        [0, 0, 1, 0, 0],
        [0, 0, 1, 0, 0],
        [0, 0, 1, 0, 0],
        [0, 0, 1, 0, 0],
        [0, 0, 0, 1, 0],
        [0, 0, 0, 1, 0],
        [0, 0, 0, 1, 0],
    ])
    CATEGORIES = 5
    """1d vector of classes (classes in [1, 2, 3]), 2d vector of onehot classes (5 possible values), and number of
    categories, respectively."""

    def _test(self, inp: np.ndarray, ref: np.ndarray, fn: Callable):
        """Checks that the given function, when the given input is supplied, behaves correctly."""
        inp = fn(inp)
        self.assertEqual(inp.shape, ref.shape)
        self.assertTrue(np.all(inp == ref), msg=f'Input:\n{inp}\nReference:\n{ref}')

    def test_classes_binary_1d(self):
        self._test(inp=self.BINARY_PROBABILITIES, ref=self.BINARY_CLASSES, fn=get_classes)

    def test_classes_binary_2d(self):
        inp = np.transpose([1 - self.BINARY_PROBABILITIES, self.BINARY_PROBABILITIES])
        self._test(inp=inp, ref=self.BINARY_CLASSES, fn=get_classes)

    def test_classes_multiclass(self):
        self._test(inp=self.MULTICLASS_PROBABILITIES, ref=self.MULTICLASS_CLASSES, fn=get_classes)

    def test_classes_multilabel(self):
        self._test(inp=self.MULTILABEL_PROBABILITIES,
                   ref=self.MULTILABEL_ONEHOT,
                   fn=lambda prob: get_classes(prob=prob, labelling=True))

    def test_onehot_binary_1d(self):
        self._test(inp=self.BINARY_CLASSES, ref=self.BINARY_CLASSES, fn=get_onehot)

    def test_onehot_binary_2d(self):
        self._test(inp=self.BINARY_CLASSES,
                   ref=self.BINARY_ONEHOT,
                   fn=lambda vector: get_onehot(vector=vector, onehot_binary=True))

    def test_onehot_multiclass(self):
        self._test(inp=self.MULTICLASS_CLASSES, ref=self.MULTICLASS_ONEHOT, fn=get_onehot)

    def test_categories_auto(self):
        with self.assertRaises(AssertionError) as context:
            self._test(inp=self.CATEGORIES_CLASSES, ref=self.CATEGORIES_ONEHOT, fn=get_onehot)
        self.assertTrue(str(context.exception).startswith('Tuples differ: (10, 4) != (10, 5)'))

    def test_categories_integer(self):
        self._test(inp=self.CATEGORIES_CLASSES,
                   ref=self.CATEGORIES_ONEHOT,
                   fn=lambda vector: get_onehot(vector=vector, classes=self.CATEGORIES))



import importlib.resources

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.model_selection import train_test_split

from moving_targets import MACS
from moving_targets.callbacks import DataLogger
from moving_targets.learners import LogisticRegression
from moving_targets.masters import ClassificationMaster
from moving_targets.metrics import Accuracy, ClassFrequenciesStd, CrossEntropy
from moving_targets.util import probabilities


# AS A FIRST STEP, WE NEED TO DEFINE OUR MASTER PROBLEM, WHICH IN THIS CASE WOULD BE THAT OF BALANCED COUNTS
class BalancedCounts(ClassificationMaster):
    def __init__(self, backend='gurobi', loss='hd', alpha='harmonic', stats=False):
        # backend : the backend, which used to solve the master step
        # loss    : the loss function, which is used to compute the master objective
        # alpha   : the alpha optimizer, which is used to balance between the gradient term and the squared term
        # stats   : whether to include master statistics in the history object

        super().__init__(backend=backend, loss=loss, alpha=alpha, stats=stats, types='discrete', labelling=False)

    # here we define the problem formulation, i.e., variables and constraints
    def build(self, x, y, p):
        # retrieve model variables from the super method, which can be either:
        #   - a 1d vector of binary variables in case of binary classification (i.e., num classes = 2)
        #   - a 2d vector of binary variables in case of multiclass classification (i.e., num classes > 2)
        # then reshape it into a column vector to better handle both cases
        variables = super(BalancedCounts, self).build(x, y, p).reshape((len(y), -1))

        # use the number of samples and classes to compute the upper bound for the number of counts of a class then
        # constraint the sum of the model variables on each column to be lower than that value
        num_samples, num_classes = (len(y), 2) if y.ndim == 1 else y.shape
        max_count = np.ceil(num_samples / num_classes)
        self.backend.add_constraints([c <= max_count for c in self.backend.sum(variables, axis=0)])

        # return the variables at the end after reshaping them into their default shape (which is the same as y)
        return variables.reshape(y.shape)


# THEN, WE MAY ADD A CUSTOM CALLBACK TO SEE HOW OUR TRAINING HAS PROCEEDED
class ClassesHistogram(DataLogger):
    def __init__(self, num_columns=4, **plt_kwargs):
        # num_columns : the number of columns to display the subplots
        # plt_kwargs  : custom arguments to be passed to the 'matplotlib.pyplot.plot' function

        super().__init__()
        self.num_columns = num_columns
        self.plt_kwargs = plt_kwargs

    def on_training_end(self, macs, x, y, p, val_data):
        # store class targets instead of class probabilities
        super(ClassesHistogram, self).on_training_end(macs, x, y, probabilities.get_classes(p), val_data)

    def on_process_end(self, macs, x, y, val_data):
        # at the end of the process, we plot the results
        plt.figure(**self.plt_kwargs)
        num_rows = int(np.ceil(len(self.iterations) / self.num_columns))
        ax = None
        # for each iteration, we we plot the classes counts for both the predictions and the adjusted targets
        for it in self.iterations:
            ax = plt.subplot(num_rows, self.num_columns, it + 1, sharex=ax, sharey=ax)
            # this check is necessary to handle the pretraining step, where no adjusted target is present
            column, name = ('y', 'targets') if it == 0 else (f'z{it}', 'adjusted')
            data = np.concatenate((self.data[column].values, self.data[f'p{it}'].values))
            hue = np.concatenate((len(self.data) * [name], len(self.data) * ['predictions']))
            sns.countplot(x=data, hue=hue, ax=ax)
            ax.set(xlabel='class', ylabel='count')
            ax.set_title(f'iteration: {it}')
        plt.show()


# FINALLY, WE CAN CREATE OUR SCRIPT TO TEST THE CORRECTNESS OF THE MODEL
if __name__ == '__main__':
    np.random.seed(0)
    sns.set_style('whitegrid')
    sns.set_context('notebook')

    # retrieve data and split train/test
    with importlib.resources.path('res', 'redwine.csv') as filepath:
        df = pd.read_csv(filepath)
        x_df, y_df = df.drop('quality', axis=1), df['quality'].astype('category').cat.codes.values
        x_tr, x_ts, y_tr, y_ts = train_test_split(x_df, y_df, stratify=y_df, shuffle=True)

    # create a moving targets instance and fit it, then plot the training history
    cbs = [ClassesHistogram(figsize=(16, 9), tight_layout=True)]
    model = MACS(
        init_step='pretraining',
        learner=LogisticRegression(x_scaler='std', max_iter=10000),
        master=BalancedCounts(),
        metrics=[Accuracy(), CrossEntropy(), ClassFrequenciesStd()]
    )
    history = model.fit(x=x_tr, y=y_tr, iterations=10, val_data={'test': (x_ts, y_ts)}, callbacks=cbs, verbose=True)
    history.plot()



import os
from typing import List, Any

import pytest

from moving_targets.masters.backends import Backend, CplexBackend
from test.masters.backends.test_backend import TestBackend


@pytest.mark.skipif(os.getenv('GITHUB_ACTIONS') == 'true', reason='No solver in Github Actions')
class TestCplexBackend(TestBackend):
    @classmethod
    def _backend(cls) -> Backend:
        return CplexBackend()

    @classmethod
    def _unsupported(cls) -> List[str]:
        return ['log', 'min', 'max', 'sqrt', 'divide', 'norm_0', 'norm_inf']

    @classmethod
    def _get_name(cls, variable: Any) -> str:
        return str(variable)



from sklearn.metrics import log_loss, precision_score, recall_score, f1_score, accuracy_score, roc_auc_score

from moving_targets.metrics import CrossEntropy, Precision, Recall, F1, Accuracy, AUC
from test.metrics.test_metrics import TestMetrics


class TestClassificationMetrics(TestMetrics):
    @staticmethod
    def _binary_generator(rng):
        y = rng.integers(0, 2, size=TestClassificationMetrics.NUM_SAMPLES)
        p = rng.random(size=TestClassificationMetrics.NUM_SAMPLES)
        return [], y, p

    @staticmethod
    def _multi_generator(rng):
        y = rng.integers(0, TestClassificationMetrics.NUM_CLASSES, size=TestClassificationMetrics.NUM_SAMPLES)
        p = rng.random(size=(TestClassificationMetrics.NUM_SAMPLES, TestClassificationMetrics.NUM_CLASSES))
        p = p / p.sum(axis=1, keepdims=1)
        return [], y, p

    def test_binary_crossentropy(self):
        self._test(data_generator=self._binary_generator,
                   mt_metric=CrossEntropy(),
                   ref_metric=lambda x, y, p: log_loss(y, p))

    def test_multi_crossentropy(self):
        self._test(data_generator=self._multi_generator,
                   mt_metric=CrossEntropy(),
                   ref_metric=lambda x, y, p: log_loss(y, p))

    def test_binary_precision(self):
        self._test(data_generator=self._binary_generator,
                   mt_metric=Precision(),
                   ref_metric=lambda x, y, p: precision_score(y, p.round().astype(int)))

    def test_multi_precision(self):
        self._test(data_generator=self._multi_generator,
                   mt_metric=Precision(average='weighted'),
                   ref_metric=lambda x, y, p: precision_score(y, p.argmax(axis=1), average='weighted'))

    def test_binary_recall(self):
        self._test(data_generator=self._binary_generator,
                   mt_metric=Recall(),
                   ref_metric=lambda x, y, p: recall_score(y, p.round().astype(int)))

    def test_multi_recall(self):
        self._test(data_generator=self._multi_generator,
                   mt_metric=Recall(average='weighted'),
                   ref_metric=lambda x, y, p: recall_score(y, p.argmax(axis=1), average='weighted'))

    def test_binary_f1(self):
        self._test(data_generator=self._binary_generator,
                   mt_metric=F1(),
                   ref_metric=lambda x, y, p: f1_score(y, p.round().astype(int)))

    def test_multi_f1(self):
        self._test(data_generator=self._multi_generator,
                   mt_metric=F1(average='weighted'),
                   ref_metric=lambda x, y, p: f1_score(y, p.argmax(axis=1), average='weighted'))

    def test_binary_accuracy(self):
        self._test(data_generator=self._binary_generator,
                   mt_metric=Accuracy(),
                   ref_metric=lambda x, y, p: accuracy_score(y, p.round().astype(int)))

    def test_multi_accuracy(self):
        self._test(data_generator=self._multi_generator,
                   mt_metric=Accuracy(),
                   ref_metric=lambda x, y, p: accuracy_score(y, p.argmax(axis=1)))

    def test_binary_auc(self):
        self._test(data_generator=self._binary_generator,
                   mt_metric=AUC(),
                   ref_metric=lambda x, y, p: roc_auc_score(y, p))

    def test_multi_auc(self):
        self._test(data_generator=self._multi_generator,
                   mt_metric=AUC(),
                   ref_metric=lambda x, y, p: roc_auc_score(y, p, multi_class='ovo'))



from typing import List, Dict, Optional, Any, Union

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import BackendError, not_implemented_message
from test.test_abstract import TestAbstract


class TestBackend(TestAbstract):
    _SIZES: Dict[str, List[int]] = {
        '1D': [TestAbstract.NUM_SAMPLES],
        '1D inv': [TestAbstract.NUM_CLASSES],
        '2D': [TestAbstract.NUM_SAMPLES, TestAbstract.NUM_CLASSES],
        '2D inv': [TestAbstract.NUM_CLASSES, TestAbstract.NUM_SAMPLES],
        '3D': [TestAbstract.NUM_SAMPLES, TestAbstract.NUM_CLASSES, TestAbstract.NUM_FEATURES],
    }
    """A dictionary which associates to each vector dimension a list representing the vector shape."""

    @classmethod
    def _backend(cls) -> Backend:
        """The `Backend` instance to be tested."""
        raise NotImplementedError(not_implemented_message(name='_backend'))

    @classmethod
    def _unsupported(cls) -> List[str]:
        """The list of unsupported operations."""
        raise NotImplementedError(not_implemented_message(name='_unsupported'))

    @classmethod
    def _get_name(cls, variable: Any) -> str:
        """Gets a formatted variable name."""
        raise NotImplementedError(not_implemented_message(name='_get_variable_name'))

    def _check_exception(self, operation: Optional[str], exception: Exception):
        """Checks if the correct exception is raised."""
        if self.__class__.__name__ == 'TestBackend':
            # if we are in the abstract test, check that the error is "NotImplementedError"
            self.assertIsInstance(exception, NotImplementedError)
        elif operation is None or operation not in self._unsupported():
            # if the operation is None or not in the unsupported,, it means that it must be supported, thus we fail
            self.fail()
        else:
            # otherwise, check that the error is a "BackendError"
            self.assertIsInstance(exception, BackendError)

    def _test_indicator_operations(self, dim: str, operation: str, **op_args):
        """Checks the correctness, using an array of the given dimensions, of the given the backend indicator operation
        that is instantiated according to the additional parameters"""
        try:
            rng = np.random.default_rng(self.SEED)
            backend = self._backend()
            sizes = (self._SIZES['1D'], 1) if dim == '0D' else (self._SIZES[dim], self._SIZES[dim])
            mt_operation = getattr(backend, operation)
            ref_operation = getattr(np, operation.lstrip('is_'))
            if operation in self._unsupported():
                backend.build()
                with self.assertRaises(BackendError):
                    a, b = [rng.random(size=size) for size in sizes]
                    a = backend.add_constants(a)
                    b = backend.add_constant(b[0]) if dim == '0D' else backend.add_constants(b)
                    mt_operation(a, b, **op_args)
                backend.clear()
            else:
                for i in range(self.NUM_TESTS):
                    # build random vectors of reference values and compute the operation result
                    ref_a, ref_b = [rng.random(size=size) for size in sizes]
                    ref_result = ref_operation(ref_a, ref_b, **op_args)
                    # build constant model variables vector(s) from values then obtain the operation result
                    backend.build()
                    mt_a = backend.add_constants(ref_a)
                    mt_b = backend.add_constant(ref_b[0]) if dim == '0D' else backend.add_constants(ref_b)
                    mt_result = mt_operation(mt_a, mt_b, **op_args)
                    backend.solve()
                    mt_result = backend.get_value(mt_result)
                    backend.clear()
                    # compare the two results
                    msg = f'Error at it. {i} with dim {dim}, {mt_result} != {ref_result}'
                    self.assertTrue(np.allclose(mt_result, ref_result, atol=10 ** -self.PLACES), msg=msg)
        except Exception as exception:
            self._check_exception(operation='indicator', exception=exception)

    def _test_numeric_operation(self, *dims: str, operation: str, **op_args):
        """Checks the correctness, using an array of the given dimensions, of the given the backend numeric operation
        that is instantiated according to the additional parameters (which must shared between the backend and numpy)"""
        try:
            rng = np.random.default_rng(self.SEED)
            backend = self._backend()
            mt_operation = getattr(backend, operation)
            ref_operation = getattr(np, operation)
            if operation in self._unsupported():
                backend.build()
                with self.assertRaises(BackendError):
                    values = [rng.random(size=self._SIZES[d]) for d in dims]
                    values = [backend.add_constants(v) for v in values]
                    mt_operation(*values, **op_args)
                backend.clear()
            else:
                for i in range(self.NUM_TESTS):
                    # build random vector(s) of reference values and compute the operation result
                    ref_values = [rng.random(size=self._SIZES[d]) for d in dims]
                    if operation == 'cov':
                        ref_result = np.cov(*ref_values, bias=True)[0, 1]
                    elif operation == 'var':
                        args = {k: v for k, v in op_args.items() if k != 'definition'}
                        ref_result = np.var(*ref_values, **args)
                    else:
                        ref_result = ref_operation(*ref_values, **op_args)
                    # build constant model variables vector(s) from values then obtain the operation result
                    backend.build()
                    mt_values = [backend.add_constants(v) for v in ref_values]
                    mt_result = mt_operation(*mt_values, **op_args)
                    backend.solve()
                    mt_result = backend.get_value(mt_result)
                    backend.clear()
                    # compare the two results
                    msg = f'Error at it. {i}, in op. {operation}, with dims {list(dims)}, {mt_result} != {ref_result}'
                    self.assertTrue(np.allclose(mt_result, ref_result, atol=10 ** -self.PLACES), msg=msg)
        except Exception as exception:
            self._check_exception(operation=operation, exception=exception)

    def _test_norm_operation(self, dim: str, norm: Union[int, str], **op_args):
        """Checks the correctness, using an array of the given dimensions, of the given the backend numeric operation
        that is instantiated according to the additional parameters (which must shared between the backend and numpy)"""
        try:
            rng = np.random.default_rng(self.SEED)
            backend = self._backend()
            mt_operation = getattr(backend, f'norm_{norm}')
            if f'norm_{norm}' in self._unsupported():
                backend.build()
                with self.assertRaises(BackendError):
                    values = rng.random(size=self._SIZES[dim])
                    values = backend.add_constants(values)
                    mt_operation(values, **op_args)
                backend.clear()
            else:
                for i in range(self.NUM_TESTS):
                    # build random vector(s) of reference values and compute the operation result
                    ref_values = rng.random(size=self._SIZES[dim])
                    if norm == 0:
                        ref_result = np.count_nonzero(ref_values, **op_args)
                    elif norm == 1:
                        ref_result = np.linalg.norm(ref_values, ord=1, **op_args)
                    elif norm == 2:
                        args = {k: v for k, v in op_args.items() if k != 'squared'}
                        ref_result = np.linalg.norm(ref_values, ord=2, **args) ** 2
                    elif norm == 'inf':
                        ref_result = np.linalg.norm(ref_values, ord=np.inf, **op_args)
                    else:
                        raise AssertionError(f"Unexpected norm {norm}")
                    # build constant model variables vector(s) from values then obtain the operation result
                    backend.build()
                    mt_values = backend.add_constants(ref_values)
                    mt_result = mt_operation(mt_values, **op_args)
                    backend.solve()
                    mt_result = backend.get_value(mt_result)
                    backend.clear()
                    # compare the two results
                    msg = f'Error at it. {i}, in norm_{norm}, with dim {dim}, {mt_result} != {ref_result}'
                    self.assertTrue(np.allclose(mt_result, ref_result, atol=10 ** -self.PLACES), msg=msg)
        except Exception as exception:
            self._check_exception(operation=f'norm_{norm}', exception=exception)

    # TEST MODELLING OPERATIONS

    def test_model(self):
        try:
            # create empty backend
            backend = self._backend()
            self.assertIsNone(backend.model)
            self.assertIsNone(backend.solution)
            # build model
            backend.build()
            self.assertIsNotNone(backend.model)
            self.assertIsNone(backend.solution)
            # solve model
            backend.solve()
            self.assertIsNotNone(backend.model)
            self.assertIsNotNone(backend.solution)
            # clear model
            backend.clear()
            self.assertIsNone(backend.model)
            self.assertIsNone(backend.solution)
        except Exception as exception:
            self._check_exception(operation=None, exception=exception)

    def test_objectives(self):
        try:
            backend = self._backend()
            ref_value = self.NUM_SAMPLES + 1
            # test minimize function and variables bounds (expected objective is -ref_value)
            backend = backend.build()
            variable = backend.add_continuous_variable(lb=-1, ub=1, name='variable')
            variables = backend.add_continuous_variables(self.NUM_SAMPLES, lb=-1, ub=1, name='variables')
            mt_value = backend.minimize(variable + backend.sum(variables)).solve().get_objective()
            self.assertAlmostEqual(mt_value, -ref_value, places=self.PLACES)
            backend.clear()
            # test maximize function and variables bounds (expected objective is +ref_value)
            backend = backend.build()
            variable = backend.add_continuous_variable(lb=-1, ub=1, name='variable')
            variables = backend.add_continuous_variables(self.NUM_SAMPLES, lb=-1, ub=1, name='variables')
            mt_value = backend.maximize(variable + backend.sum(variables)).solve().get_objective()
            self.assertAlmostEqual(mt_value, ref_value, places=self.PLACES)
            backend.clear()
        except Exception as exception:
            self._check_exception(operation='objectives', exception=exception)

    def test_variables(self):
        try:
            backend = self._backend()
            backend.build()
            for dim in ['0D', '1D', '2D', '3D']:
                if dim == '0D':
                    var = backend.add_variable(vtype='continuous', lb=0, ub=1, name='var')
                    backend.add_constraint(var >= 0, name=None)
                    backend.add_constraint(var >= 0, name='test')
                    backend.solve()
                    self.assertEqual(self._get_name(var), 'var')
                else:
                    def _expected_names(*_keys: int, _name: str = 'var') -> Union[str, List[str]]:
                        if len(_keys) == 1:
                            return [f'{_name}_{i}' for i in range(_keys[0])]
                        else:
                            res = []
                            for i in range(_keys[0]):
                                res += _expected_names(*_keys[1:], _name=f'{_name}_{i}')
                            return res

                    keys = self._SIZES[dim]
                    var = backend.add_variables(*keys, vtype='continuous', lb=0, ub=1, name='var').flatten()
                    backend.add_constraints([v >= 0 for v in var.flatten()], name=None)
                    backend.add_constraints([v >= 0 for v in var.flatten()], name='test')
                    backend.solve()
                    self.assertListEqual([self._get_name(v) for v in var], _expected_names(*keys, _name='var'))
            backend.clear()
        except Exception as exception:
            self._check_exception(operation='variables', exception=exception)

    # INDICATOR OPERATIONS

    def test_is_greater(self):
        self._test_indicator_operations('0D', operation='is_greater')
        self._test_indicator_operations('1D', operation='is_greater')
        self._test_indicator_operations('2D', operation='is_greater')
        self._test_indicator_operations('3D', operation='is_greater')

    def test_is_less(self):
        self._test_indicator_operations('0D', operation='is_less')
        self._test_indicator_operations('1D', operation='is_less')
        self._test_indicator_operations('2D', operation='is_less')
        self._test_indicator_operations('3D', operation='is_less')

    # UNARY NUMERIC OPERATIONS

    def test_sum(self):
        self._test_numeric_operation('1D', operation='sum')
        self._test_numeric_operation('2D', operation='sum')
        self._test_numeric_operation('3D', operation='sum')

    def test_sum_axis(self):
        self._test_numeric_operation('1D', operation='sum', axis=-1)
        self._test_numeric_operation('1D', operation='sum', axis=0)
        self._test_numeric_operation('2D', operation='sum', axis=-1)
        self._test_numeric_operation('2D', operation='sum', axis=0)
        self._test_numeric_operation('2D', operation='sum', axis=1)
        self._test_numeric_operation('3D', operation='sum', axis=-1)
        self._test_numeric_operation('3D', operation='sum', axis=0)
        self._test_numeric_operation('3D', operation='sum', axis=1)
        self._test_numeric_operation('3D', operation='sum', axis=2)

    def test_sqrt(self):
        self._test_numeric_operation('1D', operation='sqrt')
        self._test_numeric_operation('2D', operation='sqrt')
        self._test_numeric_operation('3D', operation='sqrt')

    def test_abs(self):
        self._test_numeric_operation('1D', operation='abs')
        self._test_numeric_operation('2D', operation='abs')
        self._test_numeric_operation('3D', operation='abs')

    def test_log(self):
        self._test_numeric_operation('1D', operation='log')
        self._test_numeric_operation('2D', operation='log')
        self._test_numeric_operation('3D', operation='log')

    def test_square(self):
        self._test_numeric_operation('1D', operation='square')
        self._test_numeric_operation('2D', operation='square')
        self._test_numeric_operation('3D', operation='square')

    def test_min(self):
        self._test_numeric_operation('1D', operation='min')
        self._test_numeric_operation('2D', operation='min')
        self._test_numeric_operation('3D', operation='min')

    def test_min_axis(self):
        self._test_numeric_operation('1D', operation='min', axis=-1)
        self._test_numeric_operation('1D', operation='min', axis=0)
        self._test_numeric_operation('2D', operation='min', axis=-1)
        self._test_numeric_operation('2D', operation='min', axis=0)
        self._test_numeric_operation('2D', operation='min', axis=1)
        self._test_numeric_operation('3D', operation='min', axis=-1)
        self._test_numeric_operation('3D', operation='min', axis=0)
        self._test_numeric_operation('3D', operation='min', axis=1)
        self._test_numeric_operation('3D', operation='min', axis=2)

    def test_max(self):
        self._test_numeric_operation('1D', operation='max')
        self._test_numeric_operation('2D', operation='max')
        self._test_numeric_operation('3D', operation='max')

    def test_max_axis(self):
        self._test_numeric_operation('1D', operation='max', axis=-1)
        self._test_numeric_operation('1D', operation='max', axis=0)
        self._test_numeric_operation('2D', operation='max', axis=-1)
        self._test_numeric_operation('2D', operation='max', axis=0)
        self._test_numeric_operation('2D', operation='max', axis=1)
        self._test_numeric_operation('3D', operation='max', axis=-1)
        self._test_numeric_operation('3D', operation='max', axis=0)
        self._test_numeric_operation('3D', operation='max', axis=1)
        self._test_numeric_operation('3D', operation='max', axis=2)

    def test_mean(self):
        self._test_numeric_operation('1D', operation='mean')
        self._test_numeric_operation('2D', operation='mean')
        self._test_numeric_operation('3D', operation='mean')

    def test_mean_axis(self):
        self._test_numeric_operation('1D', operation='mean', axis=-1)
        self._test_numeric_operation('1D', operation='mean', axis=0)
        self._test_numeric_operation('2D', operation='mean', axis=-1)
        self._test_numeric_operation('2D', operation='mean', axis=0)
        self._test_numeric_operation('2D', operation='mean', axis=1)
        self._test_numeric_operation('3D', operation='mean', axis=-1)
        self._test_numeric_operation('3D', operation='mean', axis=0)
        self._test_numeric_operation('3D', operation='mean', axis=1)
        self._test_numeric_operation('3D', operation='mean', axis=2)

    def test_var(self):
        self._test_numeric_operation('1D', operation='var', definition=True)
        self._test_numeric_operation('1D', operation='var', definition=False)
        self._test_numeric_operation('2D', operation='var', definition=True)
        self._test_numeric_operation('2D', operation='var', definition=False)
        self._test_numeric_operation('3D', operation='var', definition=True)
        self._test_numeric_operation('3D', operation='var', definition=False)

    def test_var_axis(self):
        self._test_numeric_operation('1D', operation='var', axis=-1, definition=True)
        self._test_numeric_operation('1D', operation='var', axis=-1, definition=False)
        self._test_numeric_operation('1D', operation='var', axis=0, definition=True)
        self._test_numeric_operation('1D', operation='var', axis=0, definition=False)
        self._test_numeric_operation('2D', operation='var', axis=-1, definition=True)
        self._test_numeric_operation('2D', operation='var', axis=-1, definition=False)
        self._test_numeric_operation('2D', operation='var', axis=0, definition=True)
        self._test_numeric_operation('2D', operation='var', axis=0, definition=False)
        self._test_numeric_operation('2D', operation='var', axis=1, definition=True)
        self._test_numeric_operation('2D', operation='var', axis=1, definition=False)
        self._test_numeric_operation('3D', operation='var', axis=-1, definition=True)
        self._test_numeric_operation('3D', operation='var', axis=-1, definition=False)
        self._test_numeric_operation('3D', operation='var', axis=0, definition=True)
        self._test_numeric_operation('3D', operation='var', axis=0, definition=False)
        self._test_numeric_operation('3D', operation='var', axis=1, definition=True)
        self._test_numeric_operation('3D', operation='var', axis=1, definition=False)
        self._test_numeric_operation('3D', operation='var', axis=2, definition=True)
        self._test_numeric_operation('3D', operation='var', axis=2, definition=False)

    # BINARY NUMERIC OPERATIONS

    def test_cov(self):
        self._test_numeric_operation('1D', '1D', operation='cov', definition=True)
        self._test_numeric_operation('1D', '1D', operation='cov', definition=False)

    def test_add(self):
        self._test_numeric_operation('1D', '1D', operation='add')
        self._test_numeric_operation('2D', '2D', operation='add')
        self._test_numeric_operation('3D', '3D', operation='add')

    def test_subtract(self):
        self._test_numeric_operation('1D', '1D', operation='subtract')
        self._test_numeric_operation('2D', '2D', operation='subtract')
        self._test_numeric_operation('3D', '3D', operation='subtract')

    def test_multiply(self):
        self._test_numeric_operation('1D', '1D', operation='multiply')
        self._test_numeric_operation('2D', '2D', operation='multiply')
        self._test_numeric_operation('3D', '3D', operation='multiply')

    def test_divide(self):
        self._test_numeric_operation('1D', '1D', operation='divide')
        self._test_numeric_operation('2D', '2D', operation='divide')
        self._test_numeric_operation('3D', '3D', operation='divide')

    def test_dot(self):
        self._test_numeric_operation('1D', '1D', operation='dot')
        self._test_numeric_operation('1D', '2D', operation='dot')
        self._test_numeric_operation('2D', '1D inv', operation='dot')
        self._test_numeric_operation('2D', '2D inv', operation='dot')

    def test_norm_0(self):
        self._test_norm_operation('1D', norm=0)
        self._test_norm_operation('1D', norm=0, axis=-1)
        self._test_norm_operation('1D', norm=0, axis=0)
        self._test_norm_operation('2D', norm=0, axis=-1)
        self._test_norm_operation('2D', norm=0, axis=0)
        self._test_norm_operation('2D', norm=0, axis=1)
        self._test_norm_operation('3D', norm=0, axis=-1)
        self._test_norm_operation('3D', norm=0, axis=0)
        self._test_norm_operation('3D', norm=0, axis=1)
        self._test_norm_operation('3D', norm=0, axis=2)

    def test_norm_1(self):
        self._test_norm_operation('1D', norm=1)
        self._test_norm_operation('1D', norm=1, axis=-1)
        self._test_norm_operation('1D', norm=1, axis=0)
        self._test_norm_operation('2D', norm=1, axis=-1)
        self._test_norm_operation('2D', norm=1, axis=0)
        self._test_norm_operation('2D', norm=1, axis=1)
        self._test_norm_operation('3D', norm=1, axis=-1)
        self._test_norm_operation('3D', norm=1, axis=0)
        self._test_norm_operation('3D', norm=1, axis=1)
        self._test_norm_operation('3D', norm=1, axis=2)

    def test_norm_2(self):
        self._test_norm_operation('1D', norm=2, squared=True)
        self._test_norm_operation('1D', norm=2, squared=True, axis=-1)
        self._test_norm_operation('1D', norm=2, squared=True, axis=0)
        self._test_norm_operation('2D', norm=2, squared=True, axis=-1)
        self._test_norm_operation('2D', norm=2, squared=True, axis=0)
        self._test_norm_operation('2D', norm=2, squared=True, axis=1)
        self._test_norm_operation('3D', norm=2, squared=True, axis=-1)
        self._test_norm_operation('3D', norm=2, squared=True, axis=0)
        self._test_norm_operation('3D', norm=2, squared=True, axis=1)
        self._test_norm_operation('3D', norm=2, squared=True, axis=2)

    def test_norm_inf(self):
        self._test_norm_operation('1D', norm='inf')
        self._test_norm_operation('1D', norm='inf', axis=-1)
        self._test_norm_operation('1D', norm='inf', axis=0)
        self._test_norm_operation('2D', norm='inf', axis=-1)
        self._test_norm_operation('2D', norm='inf', axis=0)
        self._test_norm_operation('2D', norm='inf', axis=1)
        self._test_norm_operation('3D', norm='inf', axis=-1)
        self._test_norm_operation('3D', norm='inf', axis=0)
        self._test_norm_operation('3D', norm='inf', axis=1)
        self._test_norm_operation('3D', norm='inf', axis=2)



from typing import Dict

import numpy as np
import pandas as pd
import sklearn
from sklearn.preprocessing import StandardScaler, MinMaxScaler, OneHotEncoder

from moving_targets.util.scalers import Scaler
from test.test_abstract import TestAbstract


class TestScalers(TestAbstract):
    """Tests the correctness of the `Scaler` class."""

    FLOAT_DTYPE = np.zeros(1, dtype='float64').dtype

    RNG = np.random.default_rng(seed=TestAbstract.SEED)

    DATA = pd.DataFrame.from_dict({
        '(0, 1)': RNG.random(TestAbstract.NUM_SAMPLES),
        '(-1, 1)': 2 * RNG.random(TestAbstract.NUM_SAMPLES) - 1,
        '{0, 2}': RNG.choice([0, 1, 2], TestAbstract.NUM_SAMPLES),
        '{M, F}': RNG.choice(['M', 'F'], TestAbstract.NUM_SAMPLES)
    })

    # decide name of 'sparse' parameter based on scikit version ('sparse' if <= 1.1, 'sparse_output' if >= 1.2)
    version = sklearn.__version__.split('.')
    kwargs = dict(sparse=False) if int(version[0]) <= 1 and int(version[1]) <= 1 else dict(sparse_output=False)

    SCALED = {
        'none': pd.DataFrame(DATA),
        'norm': pd.DataFrame(
            data=MinMaxScaler().fit_transform(DATA[['(0, 1)', '(-1, 1)', '{0, 2}']]),
            columns=['(0, 1)', '(-1, 1)', '{0, 2}']
        ),
        'std': pd.DataFrame(
            data=StandardScaler().fit_transform(DATA[['(0, 1)', '(-1, 1)', '{0, 2}']]),
            columns=['(0, 1)', '(-1, 1)', '{0, 2}']
        ),
        'onehot': pd.DataFrame(
            data=OneHotEncoder(drop='if_binary', **kwargs).fit_transform(DATA[['{0, 2}', '{M, F}']]),
            columns=['{0, 2}_0', '{0, 2}_1', '{0, 2}_2', '{M, F}_M']
        )
    }

    def _test(self, columns: Dict[str, str], scaler: Scaler, numpy: bool):
        """Checks that the given scaler, when fitted with certain data (which may be either an array or a dataframe),
        behaves as expected in the dictionary of input columns paired with the expected scaling type."""
        # handle input and reference
        one_dimensional = len(columns) == 1 and 'onehot' not in columns.values()
        input_data = self.DATA[[k for k in columns.keys()]].squeeze()
        reference_data = pd.DataFrame()
        for c, m in columns.items():
            if m == 'onehot':
                if c == '{0, 2}':
                    reference_data['{0, 2}_0'] = self.SCALED['onehot']['{0, 2}_0']
                    reference_data['{0, 2}_1'] = self.SCALED['onehot']['{0, 2}_1']
                    reference_data['{0, 2}_2'] = self.SCALED['onehot']['{0, 2}_2']
                elif c == '{M, F}':
                    reference_data['{M, F}_M'] = self.SCALED['onehot']['{M, F}_M']
                else:
                    raise AssertionError(f'Unexpected column {c}')
            else:
                reference_data[c] = self.SCALED[m][c]
        # handle types and perform dtype checks based
        if numpy:
            input_data = np.array(input_data)
            reference_data = np.array(reference_data)
            scaled_data = scaler.fit_transform(input_data)
            inverted_data = scaler.inverse_transform(scaled_data)
            self.assertIsInstance(scaled_data, np.ndarray)
            self.assertIsInstance(inverted_data, np.ndarray)
            self.assertEqual(self.FLOAT_DTYPE, scaled_data.dtype)
            self.assertEqual(input_data.dtype, inverted_data.dtype)
        else:
            scaled_data = scaler.fit_transform(input_data)
            inverted_data = scaler.inverse_transform(scaled_data)
            if one_dimensional:
                self.assertIsInstance(scaled_data, pd.Series)
                self.assertIsInstance(inverted_data, pd.Series)
                self.assertEqual(input_data.dtypes, inverted_data.dtypes)
                self.assertEqual(self.FLOAT_DTYPE, scaled_data.dtypes)
            else:
                self.assertIsInstance(scaled_data, pd.DataFrame)
                self.assertIsInstance(inverted_data, pd.DataFrame)
                self.assertDictEqual(input_data.dtypes.to_dict(), inverted_data.dtypes.to_dict())
                self.assertDictEqual({c: self.FLOAT_DTYPE for c in scaled_data.columns}, scaled_data.dtypes.to_dict())
        # check shapes
        self.assertEqual(scaled_data.shape, input_data.shape if one_dimensional else reference_data.shape)
        self.assertEqual(inverted_data.shape, input_data.shape)
        # build dataframes for compatibility
        input_data = pd.DataFrame(input_data, columns=scaler.input_columns)
        scaled_data = pd.DataFrame(scaled_data, columns=scaler.output_columns)
        inverted_data = pd.DataFrame(inverted_data, columns=scaler.input_columns)
        reference_data = pd.DataFrame(reference_data, columns=scaler.output_columns)
        # check equality between scaled and reference data, and between input and inverted data
        for c in scaler.input_columns:
            msg = str(pd.concat((input_data[c], inverted_data[c]), axis=1))
            # use try-except to deal with data types (i.e., for float-like values the check is 'all close')
            try:
                inp, inv = input_data[c].astype(np.float64), inverted_data[c].astype(np.float64)
                self.assertTrue(np.allclose(inp, inv, atol=10 ** -self.PLACES), msg=msg)
            except ValueError:
                self.assertTrue(np.array_equal(input_data[c], inverted_data[c]), msg=msg)
        for c in scaler.output_columns:
            # here there is no need to deal with data types since data should always be float-like
            msg = str(pd.concat((scaled_data[c], reference_data[c]), axis=1))
            self.assertTrue(np.allclose(scaled_data[c], reference_data[c], atol=10 ** -self.PLACES), msg=msg)

    def test_numpy_1d(self):
        self._test(columns={'(0, 1)': 'none'}, scaler=Scaler('none'), numpy=True)

    def test_pandas_1d(self):
        self._test(columns={'(0, 1)': 'none'}, scaler=Scaler('none'), numpy=False)

    def test_numpy_none(self):
        self._test(columns={'(0, 1)': 'none', '(-1, 1)': 'none', '{0, 2}': 'none'}, scaler=Scaler('none'), numpy=True)

    def test_pandas_none(self):
        self._test(columns={'(0, 1)': 'none', '(-1, 1)': 'none', '{0, 2}': 'none'}, scaler=Scaler('none'), numpy=False)

    def test_numpy_norm(self):
        self._test(columns={'(0, 1)': 'norm', '(-1, 1)': 'norm', '{0, 2}': 'norm'}, scaler=Scaler('norm'), numpy=True)

    def test_pandas_norm(self):
        self._test(columns={'(0, 1)': 'norm', '(-1, 1)': 'norm', '{0, 2}': 'norm'}, scaler=Scaler('norm'), numpy=False)

    def test_numpy_std(self):
        self._test(columns={'(0, 1)': 'std', '(-1, 1)': 'std', '{0, 2}': 'std'}, scaler=Scaler('std'), numpy=True)

    def test_pandas_std(self):
        self._test(columns={'(0, 1)': 'std', '(-1, 1)': 'std', '{0, 2}': 'std'}, scaler=Scaler('std'), numpy=False)

    def test_numpy_onehot(self):
        self._test(columns={'{0, 2}': 'onehot', '{M, F}': 'onehot'}, scaler=Scaler('onehot'), numpy=True)

    def test_pandas_onehot(self):
        self._test(columns={'{0, 2}': 'onehot', '{M, F}': 'onehot'}, scaler=Scaler('onehot'), numpy=False)

    def test_numpy_mixed(self):
        methods = {'(0, 1)': 'norm', '(-1, 1)': 'std', '{0, 2}': 'none', '{M, F}': 'onehot'}
        self._test(columns=methods, scaler=Scaler(*methods.values()), numpy=True)

    def test_pandas_mixed(self):
        methods = {'(0, 1)': 'norm', '(-1, 1)': 'std', '{M, F}': 'onehot'}
        self._test(columns={**methods, '{0, 2}': 'none'}, scaler=Scaler(default_method='none', **methods), numpy=False)



import random
from typing import Callable

import numpy as np
import sklearn.ensemble as ens
import sklearn.linear_model as lm
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import PolynomialFeatures, StandardScaler

from moving_targets.learners import Learner
from moving_targets.learners.scikit_learners import LinearRegression, LogisticRegression, RandomForestRegressor, \
    RandomForestClassifier, GradientBoostingRegressor, GradientBoostingClassifier
from test.learners.test_learners import TestLearners


class TestScikitLearners(TestLearners):

    @staticmethod
    def _custom_learner(learner_class: Callable, **kwargs) -> Learner:
        return learner_class(polynomial=3, x_scaler='std', **kwargs)

    @staticmethod
    def _custom_pipeline(model_class: Callable, **kwargs) -> Pipeline:
        return Pipeline([
            ('poly', PolynomialFeatures(degree=3, interaction_only=False, include_bias=False)),
            ('std', StandardScaler()),
            ('model', model_class(**kwargs))
        ])

    @classmethod
    def _random_state(cls):
        random.seed(cls.SEED)
        np.random.seed(cls.SEED)

    def _reference(self, learner, x, y, sample_weight: np.ndarray) -> np.ndarray:
        if isinstance(learner, Pipeline):
            learner.fit(x, y, model__sample_weight=sample_weight)
        else:
            learner.fit(x, y, sample_weight=sample_weight)
        return learner.predict_proba(x)[:, 1].squeeze() if hasattr(learner, 'predict_proba') else learner.predict(x)

    def test_linear_regression_default(self):
        self._test(mt_learner=lambda: LinearRegression(),
                   ref_learner=lambda: lm.LinearRegression(),
                   classification=False)

    def test_linear_regression_custom(self):
        self._test(mt_learner=lambda: self._custom_learner(LinearRegression),
                   ref_learner=lambda: self._custom_pipeline(lm.LinearRegression),
                   classification=False)

    def test_logistic_regression_default(self):
        self._test(mt_learner=lambda: LogisticRegression(),
                   ref_learner=lambda: lm.LogisticRegression(),
                   classification=True)

    def test_logistic_regression_custom(self):
        self._test(mt_learner=lambda: self._custom_learner(LogisticRegression),
                   ref_learner=lambda: self._custom_pipeline(lm.LogisticRegression),
                   classification=True)

    def test_random_forest_regressor_default(self):
        self._test(mt_learner=lambda: RandomForestRegressor(),
                   ref_learner=lambda: ens.RandomForestRegressor(),
                   classification=False)

    def test_random_forest_regressor_custom(self):
        self._test(mt_learner=lambda: self._custom_learner(RandomForestRegressor),
                   ref_learner=lambda: self._custom_pipeline(ens.RandomForestRegressor),
                   classification=False)

    def test_random_forest_classifier_default(self):
        self._test(mt_learner=lambda: RandomForestClassifier(),
                   ref_learner=lambda: ens.RandomForestClassifier(),
                   classification=True)

    def test_random_forest_classifier_custom(self):
        self._test(mt_learner=lambda: self._custom_learner(RandomForestClassifier),
                   ref_learner=lambda: self._custom_pipeline(ens.RandomForestClassifier),
                   classification=True)

    def test_gradient_boosting_regressor_default(self):
        self._test(mt_learner=lambda: GradientBoostingRegressor(),
                   ref_learner=lambda: ens.GradientBoostingRegressor(),
                   classification=False)

    def test_gradient_boosting_regressor_custom(self):
        self._test(mt_learner=lambda: self._custom_learner(GradientBoostingRegressor),
                   ref_learner=lambda: self._custom_pipeline(ens.GradientBoostingRegressor),
                   classification=False)

    def test_gradient_boosting_classifier_default(self):
        self._test(mt_learner=lambda: GradientBoostingClassifier(),
                   ref_learner=lambda: ens.GradientBoostingClassifier(),
                   classification=True)

    def test_gradient_boosting_classifier_custom(self):
        self._test(mt_learner=lambda: self._custom_learner(GradientBoostingClassifier),
                   ref_learner=lambda: self._custom_pipeline(ens.GradientBoostingClassifier),
                   classification=True)
